#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>

typedef void (*model_callback_t)(float t0, float *y0, float *dy);

void model(float t0, float *y0, float *dy)
{
	assert(y0 != NULL && dy != NULL);

	dy[0] = y0[1];
	dy[1] = 2.0*y0[1] - y0[0];
}

int
rk4(float h, float t0, float *y0, float *dy, float *y, unsigned int n, model_callback_t model)
{
	int i;
	float *k;/*используется для вызова модели*/
	float *k1;/*k1,..,k5 - коэффициенты в методе*/
	float *k2;
	float *k3;
	float *k4;
	float *k5;


	k1 = alloca(sizeof(float)*n);
	k2 = alloca(sizeof(float)*n);
	k3 = alloca(sizeof(float)*n);
	k4 = alloca(sizeof(float)*n);
	k5 = alloca(sizeof(float)*n);
	k = alloca(sizeof(float)*n);

	for (i = 0; i < n; i++) {
		k1[i] = h*dy[i];
		k[i] = y0[i] + 0.25*k1[i];
	}
	
	(*model)(t0+0.25*h, k, dy);
	
	for (i = 0; i < n; i++){
		k2[i] = h*dy[i];
		k[i] = y0[i] + 3.0*k1[i]/32.0 + 9.0*k2[i]/32.0;
	}
	
	(*model)(t0 + 3*h/8.0, k, dy);
	
	for (i = 0; i < n; i++) {
		k3[i] = h*dy[i];
		k[i] = y0[i] + (1932.0*k1[i] - 7200.0*k2[i] + 7296.0*k3[i])/2197.0;
	}

	(*model)(t0+12*h/13.0, k, dy);

	for (i = 0; i < n; i++) {
		k4[i] = h*dy[i];
		k[i] = y0[i] + 439.0/216.0*k1[i] - 8.0*k2[i] + 3680.0/513.0*k3[i] - 845.0/4104.0*k4[i];
	}

	(*model)(t0+h, k, dy);
	
	for (i = 0; i < n; i++) {
		k5[i] = h*dy[i];
		k[i] = y0[i] - 8.0/27.0*k1[i] + 2*k2[i] - 3544.0/2565.0*k3[i] + 1859.0/4104.0*k4[i] - 11.0/40.0*k5[i];
	}
	
	(*model)(t0+0.5*h, k, dy);

	
	for (i = 0; i < n; i++){
		y[i] = y0[i] + 25.0*k1[i]/216.0 + 1408.0/2565.0*k3[i] + 2197.0/4101.0*k4[i] - 0.2*k5[i];
	}
	for (i = 0; i < n; i++)
		y0[i] = y[i];
	return 0;
}

int rk4(float h, float t0, float *y0, float *dy, float *y, unsigned int n, model_callback_t model);
	
int main()
{
	int i;
	float h;
	float t;
	float y0[2];
	float dy[2];
	float y[2];
	unsigned int n;
	
	t = 0.0;
	h = 0.05;
	n = 2;

	/* initial values */
	y0[0] = 0;
	y0[1] = 1;
	y[0] = 0;
	y[1] = 0;

	model(t, y0, dy);

	printf("y0[0] = %e y0[1] = %e\n", y0[0], y0[1]);
	printf("dy[0] = %e dy[1] = %e\n", dy[0], dy[1]);

	for (i = 0; i < 20; i++) {
		int k;
	
		rk4(h, t, y0, dy, y, n, model);
		
		printf("t = %3.4e | y[0] = %4.3e | y[1] = %4.3e\n", t+h, y[0], y[1]);

		for (k = 0; k < n; k++)
			y0[k] = y[k];

		t += h;
	}

	return 0;
}

